/*
 * muppetcam - uploading webcam viewer
 * by muppet <scott@asofyet.org>, 20-21 june 2002
 *
 * configfile.c - read and write keys to a very simple config file
 *
 *                the format is one pair per line, key, separator, value.
 *                the current separator is a tab char.
 *                       
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <glib.h>	/* g_str* */
#include "configfile.h"

/*
 * return values of char * must be freed
 */

struct _ConfigFile {
	char       * path;
	GHashTable * data;	
};


char * config_file_path ( char * dotfile )
{
	char * homedir = g_get_home_dir ();
	if ( homedir ) {
		return g_strdup_printf ( "%s/%s", homedir, dotfile );
	} else {
		return NULL;
	}
}

ConfigFile * config_file_new ()
{
	ConfigFile * cf = (ConfigFile *) malloc (sizeof (ConfigFile));
	cf->path = NULL;
	cf->data = g_hash_table_new (g_str_hash, g_str_equal);
	return cf;
}

static void free_key_val ( char * key, char * val )
{
	if ( key ) free ( key );
	if ( val ) free ( val );
}

void config_file_destroy ( ConfigFile * cf )
{
	if ( cf->path )
		free (cf->path);
	g_hash_table_foreach (cf->data, (GHFunc)free_key_val, NULL);
	g_hash_table_destroy (cf->data);
	free (cf);
}

/*
 * this basically does
 *
 *    open IN, $filename;
 *    while (<IN>) {
 *         chomp;
 *         s/#.*$//;
 *         ($key,$val) = m/^([^\t])\t(.*)$/;
 *         $config{$key} = $val;
 *    }
 */
int config_file_read ( ConfigFile * cf, char * filename )
{
	char buf[256] = "";
	FILE * in;
	int line;

	in = fopen (filename, "r");
	if ( !in )
		return 0;

	line = 0;
	while ( ! feof(in) ) {
		char * c, * n;
		fgets(buf, 255, in);
		//fprintf(stderr, "line %d: %s", line+1, buf);
		if ( (c = strchr ( buf, '\n' )) )
			line ++;
		else if ( !feof(in) ) {
			fprintf(stderr, "%s:%d: line too long (you have "
					"found my hidden limitations!)\n",
					filename, line);
			/* probably should be line+1 there... */
		}
		/* chop off everything after the comment character... */ 
		if ( (c = strchr ( buf, '#' )) )
			*c = '\0';
		
		/* ignore blank lines. */
		c = buf;
		while ( *c && isspace(*c) )
			c++;
		if ( ! *c )
			continue;
		
		/* parse key\tval */
		if ( (c = strchr ( buf, '\t')) ) {
			*c++ = '\0';
			/* zap trailing newline */
			if ( (n = strrchr ( c, '\n' )) )
				*n = '\0';
			config_file_value_set ( cf, buf, c );
			//fprintf(stderr,"found '%s' => '%s'\n", buf, c);
		} else {
			//fprintf(stderr, "%s:%d: bad format (no tab character"
			//		" between key and value)\n",
			//		filename, line);
		}
	}
	
	fclose (in);
	return 1;
}

static void write_key ( char * key, char * val, FILE * out )
{
	/* write key and val */
	fprintf ( out, "%s\t%s\n", key, val );
}

int config_file_write ( ConfigFile * cf, char * filename )
{
	FILE * out;
	out = fopen (filename, "w");
	if ( ! out ) {
		perror (filename);
		return 0;
	}
	/* write header */
	g_hash_table_foreach (cf->data, (GHFunc)write_key, out);
	fclose (out);
	return 1;
}

char * config_file_value_get ( ConfigFile * cf, char * key )
{
	char * val = (char*) g_hash_table_lookup (cf->data, key );
	if ( val )
		/* return a copy */
		return g_strdup ( val );
	else
		return NULL;
}

void config_file_value_set ( ConfigFile * cf, char * key, char * value )
{
	char * val;
	/* don't put in a NULL value for a valid key! */
	if ( ! value )
		value = "";
	val = (char*) g_hash_table_lookup ( cf->data, key );
	if ( val ) {
		/* already exists.  free val, reuse key. */
		free ( val );
		g_hash_table_insert ( cf->data, 
				key, 
				g_strdup(value) );
	} else {
		/* doesn't already exist.  dupe both. */
		g_hash_table_insert ( cf->data, 
				g_strdup(key), 
				g_strdup(value) );
	}
}

int  config_file_value_get_double ( ConfigFile * cf, char * key, double * value )
{
	char * strval;
	strval = config_file_value_get ( cf, key );
	if ( ! strval )
		return 0;
	*value = atof ( strval );
	free ( strval );
	return 1;
}

int  config_file_value_get_int   ( ConfigFile * cf, char * key, int * value )
{
	char * strval;
	strval = config_file_value_get ( cf, key );
	if ( ! strval )
		return 0;
	*value = atoi ( strval );
	free ( strval );
	return 1;
}

void config_file_value_set_double ( ConfigFile * cf, char * key, double value )
{
	char buf[256];
	g_snprintf ( buf, 255, "%f", value );
	config_file_value_set ( cf, key, buf );
}

void config_file_value_set_int   ( ConfigFile * cf, char * key, int value )
{
	char buf[256];
	g_snprintf ( buf, 255, "%d", value );
	config_file_value_set ( cf, key, buf );
}


#if 0

#define DOTFILE	".muppetcam"

int main ( int argc, char * argv [] )
{
	ConfigFile * cf;
	char * filename = config_file_path ( DOTFILE );

	if ( ! filename ) {
		perror ( "couldn't find homedir" );
		exit ( -1 );
	}

	puts(filename);

	cf = config_file_new ();

	config_file_read (cf, filename);
//	config_file_value_set (cf, "one", "the first key");
//	config_file_value_set (cf, "two", "the second key");
	
	config_file_write (cf, filename);
	
	config_file_destroy (cf);
	
	free ( filename );

	return 1;
}

#endif
