/*
 * muppetcam - uploading webcam viewer
 * by muppet <scott@asofyet.org>, 20-21 june 2002
 *
 * works with Video4Linux devices and the 3Com HomeConnect USB camera (using 
 * the vicam code from http://homeconnectusb.sourceforge.net)
 *
 * muppetcam.c - main() and gui
 */
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <sys/types.h>	/* pid_t */
#include <unistd.h>	/* fork and _exit */
#include <wait.h>	/* waitpid */
#include <signal.h>	/* sigaction */
#include <time.h>	/* for ctime_r, et al */

#include "camera.h"
#include "vicam.h"
#include "v4lcam.h"
#include "muppetcam.h"
#include "serverdialog.h"
#include "configfile.h"
#include "viutility.h"		/* vicam usb stuff for main() */

/* global configuration info */
ConfigFile * config;

/*****************************************************************************/

/* callbacks */

/* for the image canvas */
static int configure_event            ( GtkWidget         * canvas, 
                                        GdkEventConfigure * event, 
                                        MuppetCamWindow   * mc );
static int expose_event               ( GtkWidget         * canvas, 
                                        GdkEventExpose    * event, 
                                        MuppetCamWindow   * mc );

/* for widgets in the main window */
static void live_toggled              ( GtkWidget         * button, 
                                        MuppetCamWindow   * mc );
static void snap_clicked              ( GtkWidget         * button, 
                                        MuppetCamWindow   * mc );
#ifdef DOSAVE
static void save_clicked              ( GtkWidget         * button,
                                        MuppetCamWindow   * mc );
#endif
static void freq_spin_changed         ( GtkWidget         * spin, 
                                        MuppetCamWindow   * mc );
static void upload_toggled            ( GtkWidget         * check, 
                                        MuppetCamWindow   * mc );

/* callbacks for the annotations */
static int annotation_entry_changed   ( GtkWidget         * entry,  
                                        Annotation        * a );
static int annotation_set_clicked     ( GtkWidget         * button,
                                        Annotation        * a );
static int annotation_reset_clicked   ( GtkWidget         * button, 
                                        Annotation        * a );

/* "live" video timeout function */
static int  livish_timeout            ( MuppetCamWindow   * mc );


/*****************************************************************************/

/*
 * annotation
 *    called annotation because it's for the text strings used as image 
 *    annotations, it really gives you a string that you can set or reset 
 *    explicitly.  this avoids the need for a dialog, so that the strings 
 *    are always visible on the main window.
 *    
 *    in a show of bad form, you allocate the memory elsewhere, and this
 *    fills it.
 *
 *    returns the toplevel widget so you can pack the whole thing.
 */
GtkWidget * create_annotation ( Annotation * a, char * text )
{
	a->hbox = gtk_hbox_new ( FALSE, 0 );
	a->entry = gtk_entry_new ();
	a->set = gtk_button_new_with_label ( "set" );
	a->reset = gtk_button_new_with_label ( "reset" );
	gtk_box_pack_start ( GTK_BOX(a->hbox), a->entry, TRUE, TRUE, 0 );
	gtk_box_pack_start ( GTK_BOX(a->hbox), a->set, FALSE, FALSE, 0 );
	gtk_box_pack_start ( GTK_BOX(a->hbox), a->reset, FALSE, FALSE, 0 );
	a->text = g_strdup ( text );
	gtk_entry_set_text ( GTK_ENTRY(a->entry), a->text );
	gtk_widget_set_sensitive ( a->set, 0 );
	gtk_widget_set_sensitive ( a->reset, 0 );
	gtk_signal_connect ( GTK_OBJECT(a->entry), "changed",
			GTK_SIGNAL_FUNC(annotation_entry_changed), a );
	gtk_signal_connect ( GTK_OBJECT(a->set), "clicked",
			GTK_SIGNAL_FUNC(annotation_set_clicked), a );
	gtk_signal_connect ( GTK_OBJECT(a->reset), "clicked",
			GTK_SIGNAL_FUNC(annotation_reset_clicked), a );
	return a->hbox;
}

int annotation_entry_changed ( GtkWidget * entry, Annotation * a )
{
	gtk_widget_set_sensitive ( a->set, 1 );
	gtk_widget_set_sensitive ( a->reset, 1 );
	return 1;
}

int annotation_set_clicked ( GtkWidget * button, Annotation * a )
{
	free ( a->text );
	a->text = g_strdup ( gtk_entry_get_text ( GTK_ENTRY( a->entry ) ) );
	gtk_widget_set_sensitive ( a->set, 0 );
	gtk_widget_set_sensitive ( a->reset, 0 );
	return 1;
}

int annotation_reset_clicked ( GtkWidget * button, Annotation * a )
{
	gtk_entry_set_text ( GTK_ENTRY(a->entry), a->text );
	gtk_widget_set_sensitive ( a->set, 0 );
	gtk_widget_set_sensitive ( a->reset, 0 );
	return 1;
}




/*
 * duh, create and show the application's main window.  
 * live video starts out disabled.
 */
MuppetCamWindow * create_and_show_main_window ( Camera * cam )
{
	MuppetCamWindow * mc;
	GtkWidget * vbox;
	GtkWidget * hbox;
	GtkWidget * label;
	GtkAdjustment * adj;
	GtkWidget * spin;
	GtkWidget * button;
	int i;
	char * text;

	mc = (MuppetCamWindow*) calloc ( 1, sizeof(MuppetCamWindow) );

	mc->cam = cam;

	/* grab some values from config */
	if ( ! config_file_value_get_int ( config, 
				"upload_freq", &mc->upload_freq ) )
		mc->upload_freq = DEFAULT_UPLOAD_FREQ;

	mc->hostname = config_file_value_get ( config, "hostname" );
	mc->filename = config_file_value_get ( config, "filename" );

	/*
	 * start creating widgets.  just pack them, and we'll do a 
	 * show_all on the toplevel when we're finished.
	 */
	mc->window = gtk_window_new ( GTK_WINDOW_TOPLEVEL );
	gtk_window_set_title ( GTK_WINDOW(mc->window), "MuppetCam" );
	gtk_signal_connect ( GTK_OBJECT(mc->window), "delete-event",
			GTK_SIGNAL_FUNC(gtk_main_quit), NULL );
	
	vbox = gtk_vbox_new ( FALSE, 0 );
	gtk_container_add ( GTK_CONTAINER(mc->window), vbox );

	mc->canvas = gtk_drawing_area_new ();
	gtk_drawing_area_size ( GTK_DRAWING_AREA(mc->canvas),
			mc->cam->width, mc->cam->height );
	gtk_box_pack_start ( GTK_BOX(vbox), mc->canvas, TRUE, TRUE, 0 );

	gtk_signal_connect ( GTK_OBJECT(mc->canvas), "configure_event",
			GTK_SIGNAL_FUNC(configure_event), mc );
	gtk_signal_connect ( GTK_OBJECT(mc->canvas), "expose_event",
			GTK_SIGNAL_FUNC(expose_event), mc );

	
	/* command buttons */
	hbox = gtk_hbox_new ( FALSE, 0 );
	gtk_box_pack_start ( GTK_BOX(vbox), hbox, FALSE, FALSE, 0 );

	mc->live = gtk_toggle_button_new_with_label ( "Live" );
	gtk_box_pack_start ( GTK_BOX(hbox), mc->live, FALSE, FALSE, 0 );
	gtk_signal_connect ( GTK_OBJECT(mc->live), "toggled", 
			GTK_SIGNAL_FUNC(live_toggled), mc );
	gtk_toggle_button_set_active ( GTK_TOGGLE_BUTTON(mc->live), 0 );
	
	mc->snap = gtk_button_new_with_label ( "Snap" );
	gtk_box_pack_start ( GTK_BOX(hbox), mc->snap, FALSE, FALSE, 0 );
	gtk_signal_connect ( GTK_OBJECT(mc->snap), "clicked", 
			GTK_SIGNAL_FUNC(snap_clicked), mc );
	
#ifdef DOSAVE
	mc->save = gtk_button_new_with_label ( "Save" );
	gtk_box_pack_start ( GTK_BOX(hbox), mc->save, FALSE, FALSE, 0 );
	gtk_signal_connect ( GTK_OBJECT(mc->save), "clicked", 
			GTK_SIGNAL_FUNC(save_clicked), mc );
	gtk_widget_set_sensitive ( mc->save, 0 );
#endif


	mc->status = gtk_label_new ( "muppetcam ready for action" );
	gtk_box_pack_start ( GTK_BOX(hbox), mc->status, FALSE, FALSE, 3 );
	
	
	/* camera controls */
	{
	GtkWidget * frame = gtk_frame_new ( cam->name );
	GtkWidget * vbox2 = gtk_vbox_new ( FALSE, 0 );
	gtk_box_pack_start ( GTK_BOX(vbox), frame, FALSE, FALSE, 0 );
	gtk_container_add ( GTK_CONTAINER(frame), vbox2 );
	gtk_container_set_border_width ( GTK_CONTAINER(frame), 3 );
	gtk_container_set_border_width ( GTK_CONTAINER(vbox2), 3 );
	for ( i = 0 ; i < mc->cam->ncontrols ; i++ )
		//gtk_box_pack_start ( GTK_BOX(vbox), 
		gtk_box_pack_start ( GTK_BOX(vbox2), 
			cam_control_create_widget ( mc->cam->control[i] ),
			TRUE, TRUE, 0 );
	}
	
	/* upload controls 
	 *
	 *   check      label  spin  label      button
	 * [x] upload   every [ n |] frames    [ to server... ] 
	 */
	hbox = gtk_hbox_new ( FALSE, 0 );
	gtk_box_pack_start ( GTK_BOX(vbox), hbox, FALSE, FALSE, 0 );

	mc->upload = gtk_check_button_new_with_label ( "upload" );
	gtk_box_pack_start ( GTK_BOX(hbox), mc->upload, FALSE, FALSE, 0 );
	mc->upload_frame = gtk_hbox_new ( FALSE, 0 );
	gtk_box_pack_start ( GTK_BOX(hbox), mc->upload_frame, TRUE, TRUE, 0 );
	label = gtk_label_new ( "every " );
	gtk_box_pack_start ( GTK_BOX(mc->upload_frame), label, FALSE, FALSE, 0 );
	adj = (GtkAdjustment*) gtk_adjustment_new 
			( mc->upload_freq, 1, 3600, 1, 10, 10 );
	spin = gtk_spin_button_new ( adj, 1.0, 0 );
	gtk_box_pack_start ( GTK_BOX(mc->upload_frame), spin, FALSE, FALSE, 0 );
	label = gtk_label_new ( " frames" );
	gtk_box_pack_start ( GTK_BOX(mc->upload_frame), label, FALSE, FALSE, 0 );

	gtk_toggle_button_set_active ( GTK_TOGGLE_BUTTON(mc->upload), 0 );
	gtk_widget_set_sensitive ( mc->upload_frame, 0 );
	
	gtk_signal_connect ( GTK_OBJECT(spin), "changed", 
			GTK_SIGNAL_FUNC(freq_spin_changed), mc );
	gtk_signal_connect ( GTK_OBJECT(mc->upload), "toggled", 
			GTK_SIGNAL_FUNC(upload_toggled), mc );

	button = gtk_button_new_with_label ( "to server(s)..." );
	gtk_box_pack_start ( GTK_BOX(hbox), button, FALSE, FALSE, 0 );

	gtk_signal_connect ( GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(server_clicked), mc );
	
	/* image annotation entries */
	text = config_file_value_get ( config, "toptext" );
	gtk_box_pack_start ( GTK_BOX(vbox), 
			create_annotation 
				( &(mc->toptext), (text ? text : "") ),
			FALSE, FALSE, 0 );
	if ( text ) free ( text );
	
	text = config_file_value_get ( config, "bottomtext" );
	gtk_box_pack_start ( GTK_BOX(vbox), 
			create_annotation 
				( &(mc->bottomtext), (text ? text : "") ),
			FALSE, FALSE, 0 );
	if ( text ) free ( text );
	
	gtk_widget_show_all ( mc->window );

	return mc;
}

/*
 * some callbacks for the widgets in the main window... 
 */

void freq_spin_changed ( GtkWidget * spin, MuppetCamWindow * mc )
{
	mc->upload_freq = 
		gtk_spin_button_get_value_as_int ( GTK_SPIN_BUTTON(spin) );
}

void upload_toggled ( GtkWidget * check, MuppetCamWindow * mc )
{
	/* if upload is turned off, disable the associated controls */
	gtk_widget_set_sensitive ( mc->upload_frame,
		gtk_toggle_button_get_active ( GTK_TOGGLE_BUTTON(check) ) );
}


/* 
 * for the image display widget.
 * create (or resize) the backing pixmap, and initialize it with a fresh
 * shot from the camera.
 */
int configure_event ( GtkWidget         * canvas, 
                      GdkEventConfigure * event, 
                      MuppetCamWindow   * mc )
{
	if ( mc->pixmap )
		gdk_pixmap_unref ( mc->pixmap );

	mc->pixmap = gdk_pixmap_new ( canvas->window,
	                              canvas->allocation.width,
	                              canvas->allocation.height,
				      -1 );

	/* like in the snap button callback, snap twice to make sure the new 
	 * settings get flushed to the camera. */
	snap ( mc );
	snap ( mc );
	draw ( mc );
	return TRUE;
}

/*
 * again for the image display widget. 
 */
int expose_event ( GtkWidget       * canvas, 
                   GdkEventExpose  * event, 
                   MuppetCamWindow * mc )
{
	gdk_draw_pixmap ( canvas->window,
			canvas->style->fg_gc[GTK_WIDGET_STATE (canvas)],
			mc->pixmap,
			event->area.x, event->area.y,
			event->area.x, event->area.y,
			event->area.width, event->area.height );

	return FALSE;
}


/*
 * the "live" video timeout function.  this gets called every UPDATE_SPEED
 * milliseconds (when the "live" button is on) to show a new frame on the 
 * screen.
 *
 * note that this snaps only once, so any new camera settings will be sent
 * on this frame but used on the next one.
 */
int livish_timeout ( MuppetCamWindow * mc )
{
	snap ( mc );
	draw ( mc );

	/* the rest is only of concern if we are uploading frames. */
	if ( ! gtk_toggle_button_get_active ( GTK_TOGGLE_BUTTON(mc->upload) ) )
		return 1;
	
	mc->current++;
	if ( mc->current >= mc->upload_freq ) {
		set_status ( mc, "uploading to %s", mc->hostname );
		mc->current = 0;
		upload ( mc );
		
	} else {
		set_status ( mc, "upload in %d", 
				mc->upload_freq - mc->current );
	}
	return 1;
}


/*
 * activate or deactivate the "live video" timeout.
 */
void live_toggled ( GtkWidget * button, MuppetCamWindow * mc )
{
	if ( ! gtk_toggle_button_get_active ( GTK_TOGGLE_BUTTON(button) ) ) {
		gtk_timeout_remove ( mc->timeoutid );
		gtk_widget_set_sensitive ( mc->snap, 1 );
#ifdef DOSAVE
		gtk_widget_set_sensitive ( mc->save, 1 );
#endif
	} else {
		mc->timeoutid = gtk_timeout_add ( UPDATE_SPEED, 
					(GtkFunction)livish_timeout,
					mc );
		gtk_widget_set_sensitive ( mc->snap, 0 );
#ifdef DOSAVE
		gtk_widget_set_sensitive ( mc->save, 0 );
#endif
	}
}

/*
 * single-shot.
 * shouldn't get here unless live video is off (because the button is 
 * supposed to be disabled when live video is on).
 */
void snap_clicked ( GtkWidget * button, MuppetCamWindow * mc )
{
	/* capture twice to make sure shutter setting gets applied */
	snap ( mc );
	snap ( mc );
	draw ( mc );
}

/*
void save_clicked ( GtkWidget * button, MuppetCamWindow * mc )
{
}
*/


/*
 * set the message in the status line.  works like printf.
 */
void set_status ( MuppetCamWindow * mc, const char *fmt, ... )
{
	char * s;

	va_list ap;
	va_start(ap, fmt);
	s = g_strdup_vprintf(fmt, ap);
	if ( ! s ) {
		/*
		 * this could happen in intense memory situations.
		 * rather than just toss the message, force it to
		 * stderr.
		 */
		perror("set_status: g_strdup_vprintf failed");
		fprintf(stderr,"message was: ");
		vfprintf(stderr,fmt,ap);
		fprintf(stderr,"\n");
		va_end(ap);
	} else {
		va_end(ap);
		gtk_label_set_text ( GTK_LABEL(mc->status), s);
		free(s);
	}
}



void sigchld_handler ( int n )
{
	int ret;
	int status;
	do {
		/* reap any and all finished child processes. */
		ret = waitpid ( -1, &status, WNOHANG );
		if ( ret < 0 )
			perror ( "waitpid" );

	} while ( ret > 0 );
}



/*
 * attempt to upload a frame to the configured destination server(s).
 * 
 * we use external programs for the image annotation and transfer (since 
 * they're far better at it than my code could be).
 * 
 * to avoid locking up the gui while they run (my ftp is notorious for taking
 * a long time while negotiating a security setting), we'll fork a child 
 * process to do that work.
 *
 * NOTE: i'm using _exit() to quit the other process to avoid that other
 *       process trying to close gtk's sockets on exit.  is that really a
 *       problem?
 *
 * TODO - unhardcode sizes
 * TODO - use Gdk to draw text in arbitrary user-selectable fonts, and then
 *        grab that GdkPixmap back from the server?  might same some time, and
 *        make for cool points by selecting the font and color of your choice.
 * TODO - make the ftp script configurable
 * TODO - this currently will just bomb if convert and ftp aren't in your 
 *        path.  what if the user needs to use a different program?
 * TODO - look at using a thread instead of a process, so we can communicate
 *        errors back to the UI.
 * TODO - what about multiple upload servers?
 */
void upload ( MuppetCamWindow * mc )
{
	FILE * out;
	char * cmd;
	char timebuf[21];
	time_t now;
	pid_t pid;

	debug("checking hostname and filename settings\n");
	if ( ! ( mc->hostname && mc->filename ) ) {
		set_status ( mc, "invalid server settings" );
		return;
	}

	/* 
	 * write the image file before forking to avoid any race conditions
	 * on the image buffer (you know, copy-on-write forking, by the time
	 * we try to access it, it might have been changed).
	 */
	debug("writing raw file before fork to avoid race on buffer\n");
	out = fopen ( "snap.raw", "wb" );
	if ( ! out ) {
		/* ack!  let 'em know what happened, and disable the upload
		 * function so they can fix it and reenable. */
		set_status ( mc, "snap.raw: %s", strerror(errno) );
		gtk_signal_emit_by_name ( GTK_OBJECT(mc->upload), "clicked" );
		return;
	}
	fwrite ( mc->cam->image, mc->cam->depth, 
			mc->cam->width * mc->cam->height, out );
	fclose ( out );
	
	debug("forking\n");
	pid = fork ();
	if ( pid < 0 ) {
		/* eeek! */
		perror("couldn't fork");
		return;
	} else if ( pid > 0 ) {
		struct sigaction action;
		/* parent... */
		debug("we are in the parent, child is %d.\n", pid);

		action.sa_handler = sigchld_handler;
		action.sa_flags = SA_ONESHOT;
		/* block all signals during the handler */
		sigfillset(&action.sa_mask);
		
		debug("installing reaper.\n");
		sigaction ( SIGCHLD, &action, NULL );

//		/* let the user know we've started an upload */
//		gdk_beep();
		
		/* now just go back to what we were doing. */
		return;
	}
	
	/* we are in the child. */

	/* first, annotate the image with the user's top and bottom text.
	 * we'll use imagemagick's convert to do that.  convert can read
	 * the drawing commands from a text file, which is a little more
	 * robust at handling funny characters in the text strings than
	 * trying to escape them from a shell, so let's do that. */
	
	debug("child: alive, writing command file\n");

	/* automagickally append a timestamp to the bottom text */
	now = time ( NULL );
	ctime_r ( &now, timebuf );
	timebuf[19] = '\0';
	timebuf[20] = '\0';

/* FIXME -- hardcoded image size and text location is uncool */
	
	out = fopen ( "comment", "w" );
	if ( ! out ) {
		perror ( "comment (convert drawing command file)" );
		_exit(1);
	}
	fprintf ( out, "text 1,10 \"%s\"", mc->toptext.text );
	fprintf ( out, "text 1,239 \"%s %s\"", mc->bottomtext.text, timebuf );
	fclose ( out );

	/* this is pretty much the command suggested in the vigrab readme for
	 * converting the vigrab raw images from 512x242 to something normal,
	 * except that i already have a 320x240 image.  in particular, is the
	 * sharpen still necessary? */
	cmd = g_strdup_printf ( "convert -depth 8 -size 320x240 rgb:snap.raw "
			"-geometry 320x240! -sharpen 5x2 "
			"-fill white -box black -draw \"@comment\" "
			"\"%s\"", "snap.jpg" );

	debug("child: cmd is \"%s\"\n", cmd);
	system ( cmd );

//	debug("child: free()ing cmd\n");
	free ( cmd );

	/* now transfer the image to the server */
	cmd = g_strdup_printf ( "ftp <<EOF\n"
				"open %s\n"
				"bin\n"
				/*"ls\n"*/
				/*"hash\n"*/
				"put snap.jpg \"%s\"\n"
				"bye\n"
				"EOF\n",
			mc->hostname, mc->filename );

	debug("child: cmd is \"%s\"\n", cmd);
	system ( cmd );

//	debug("child: free()ing cmd\n");
	free ( cmd );

	debug("child: calling _exit()\n");
	_exit(0);
}

/*
 * snap an image from the camera...
 */
int snap ( MuppetCamWindow * mc )
{
	return camera_snap ( mc->cam );
}

/*
 * draw the cam's current image data onto the backing pixmap (and flush to
 * the screen).
 */
int draw ( MuppetCamWindow * mc )
{
	GdkRectangle update_rect;

	if ( camera_is_color ( mc->cam ) ) {
		gdk_draw_rgb_image ( mc->pixmap, 
				mc->canvas->style->fg_gc[GTK_STATE_NORMAL], 
				0, 0, mc->cam->width, mc->cam->height,
				GDK_RGB_DITHER_NONE,
				mc->cam->image, 
				mc->cam->width*mc->cam->depth);
	} else {
		gdk_draw_gray_image ( mc->pixmap, 
				mc->canvas->style->fg_gc[GTK_STATE_NORMAL], 
				0, 0, mc->cam->width, mc->cam->height,
				GDK_RGB_DITHER_NONE,
				mc->cam->image, 
				mc->cam->width*mc->cam->depth);
	}

	update_rect.x = 0;
	update_rect.y = 0;
	update_rect.width = mc->cam->width;
	update_rect.height = mc->cam->height;
	gtk_widget_draw ( mc->canvas, &update_rect );
	
	return 1;
}


static void usage ( char * progname )
{
	printf (
"MuppetCam " VERSION " (compiled "__DATE__ ")\n"
"  written by the muppet, scott@asofyet.org\n"
"  uploading webcam GUI\n"
"  usage:  %s [options] [device]\n"
"  options:\n"
"    -v  drive a ViCam / 3Com HomeConnectUSB camera without the kernel driver.\n"
"        the vicam/3comhc kernel driver is still a little beta; using this\n"
"        option will give you slightly better image quality, but requires\n"
"        that the kernel module is *not* loaded, and that the program runs\n"
"        as root (for direct access to the USB hardware).\n"
"        note: Gtk+ will not run setuid.  this option uses code from the guys\n"
"        at http://homeconnectusb.sourceforge.net .\n"
"    -c PATH\n"
"        use config file PATH instead of ~/.muppetcam\n"
"    -h  show this message\n"
"    device\n"
"        the device file of the Video4Linux device to use.  the default is\n"
"        /dev/video0.  (currently tested only with the two cameras i have.)\n"
"\n",
		progname );
}


/*
 *
 *
 */
int main ( int argc, char * argv[] )
{
	char c;
	int i;
	int cams;		/* number of USB devices found (vicam only) */
	MuppetCamWindow * mc;	/* the big cahuna */
	Camera * cam;		/* all-important camera object */
	int use_vicam = 0;
	char * prefsfile = NULL;
	char * v4ldevice = "/dev/video0";

	/*
	 * we have to initialize the gtk type system before we can create any
	 * Camera objects, and since gtk won't run setuid, it might be a good
	 * idea to check for that here before we bother doing anything else.
	 */
	gtk_set_locale ();
	gtk_init ( & argc, & argv );
	gdk_rgb_init ();
	
	while ( -1 < (c = getopt (argc, argv, "vc:h")) ) {
		switch ( c ) {
		    case 'v':
			use_vicam = 1;
			break;
		    case 'c':
			/* prefsfile gets free()d below! */
			prefsfile = g_strdup ( optarg );
			whine ( "using preferences file %s\n", prefsfile );
			break;
		    case 'h':
			usage ( g_basename ( argv[0] ) );
			/* that leaks, but it doesn't matter, because... */
			exit(0);
		    default:
			whine ( "ignoring unknown option %c\n"
				"try %s -h for help\n", c, argv[0] );
			break;
		}
	}
	if ( optind < argc )
		v4ldevice = argv[optind];

	/* 
	 * initialize config object.
	 * it's quite alright if the preferences file doesn't exist;
	 * we'll just use the defaults.
	 */
	if ( ! prefsfile )
		prefsfile = config_file_path ( ".muppetcam" );
	config = config_file_new ();
	config_file_read (config, prefsfile);

	/*
	 * now select a device and create the Camera object.
	 */
	
	if ( use_vicam /* || list all available */ ) {
		/* intialize the usb library and look for any vicams. */
		cams = vicam_usb_init ();
		if ( cams < 0 ) {
			perror ( "vicam_usb_init" );
			exit ( 1 );
		} else if ( 0 == cams ) {
			printf ( "No HomeConnectUSB cameras found!\n" );
			exit ( 2 );
		}
	}

	if ( /* list all available || */ !use_vicam  ) {
		/* TODO scan /dev/video(n) for valid cameras */
	}

	/* TODO - if the user hasn't already selected the camera on the 
	 *        command line, present a choice of devices. */

	if ( use_vicam ) {
		/* use a vicam */
		/* only works with one camera right now... (basing this off 
		 * of the vicam tools from homeconnectusb.sourceforge.net, 
		 * so i do as they do.) */
		cam = vicam_new ( 0 );
		
	} else {
		/* use a V4L device */
		cam = v4lcam_new ( v4ldevice );
	}

	if ( ! cam ) {
		whine("couldn't open the camera, so there's nothing to do!\n");
		exit(-1);
	}
	
	/* apply the user's preferred the camera control values */
	for ( i = 0 ; i < cam->ncontrols ; i++ ) {
		int val;
		if ( config_file_value_get_int ( config,
				camera_control_name(cam, i), &val ) )
			gtk_adjustment_set_value ( cam->control[i]->adj, 
					(gfloat) val );
	}

	/* now we're ready to get started on the gui... */
	mc = create_and_show_main_window ( cam );

	/* spin here, watching the paint dry... */
	gtk_main ();

	/*
	 * save the user's settings for next time.
	 */
	for ( i = 0 ; i < cam->ncontrols ; i++ )
		config_file_value_set_int ( config,
				camera_control_name(cam, i),
				camera_control_value(cam, i) );
	config_file_value_set ( config, "toptext", mc->toptext.text );
	config_file_value_set ( config, "bottomtext", mc->bottomtext.text );
	config_file_value_set ( config, "hostname", mc->hostname );
	config_file_value_set ( config, "filename", mc->filename );
	
	config_file_write ( config, prefsfile );

	/* clean up */
	free ( prefsfile );
	config_file_destroy ( config );
	camera_destroy ( cam );

	/* fin */
	
	return 0;
}
