/*
 * muppetcam - uploading webcam viewer
 * by muppet <scott@asofyet.org>, 20-21 june 2002
 *
 * v4lcam.c - MuppetCam driver for Video4Linux devices.
 *            This is a subclass of Camera.
 *
 *            This reads the camera in normal read mode, NOT mmap, as i've
 *            had rather poor experiences with mmap in the past.
 *
 *            I've decided to go head and handle various known camera 
 *            drivers' quirks in here, so that they are transparent to
 *            the next layer.  This may prove to complicate things, but
 *            it looks okay for now.
 *
 *            Currently tested with 
 *               Model                          Driver Module
 *               ============================================
 *               Kensington VideoCAM 6701       se401.o
 *               3Com HomeConnect WebCam        3comhc.o (NOT vicam.o)
 */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/videodev.h>	/* for the video device stuff here... */

#include "camera.h"
#include "v4lcam.h"

/* we don't really have anything extra to carry at this point... */
typedef struct {
	Camera   cam;
} V4LCam;


static int  v4lcam_snap    ( Camera * cam );
static void v4lcam_destroy ( Camera * cam );


Camera * v4lcam_new ( char * device )
{
	V4LCam                  * v4lcam;       /* the camera object (duh)  */
	int                       fd;           /* device file descriptor   */
	struct video_capability   cap;          /* vid caps struct          */
	struct video_window       win;          /* vid window struct        */
	struct video_picture      pic;          /* picture desc struct      */

	/* "you know i cain't play without no geetar strangs... 
	 * heeeere kitty kitty!" */
	fd = open ( device, O_RDONLY );
	if ( fd < 0 ) {
		perror ( device );
		return NULL;
	}

	/* find out a little about our new friend. */
	if ( 0 > ioctl ( fd, VIDIOCGCAP, & cap ) ) {
		perror("VIDIOCGCAP");
		return NULL;
	}
	if ( 0 > ioctl ( fd, VIDIOCGWIN, & win ) ) {
		perror("VIDIOCGWIN");
		return NULL;
	}
	if ( 0 > ioctl ( fd, VIDIOCGPICT, & pic ) ) {
		perror("VIDIOCGPIC");
		return NULL;
	}

	/* 
	 * catch known camera types for any special things we need to do. 
	 */
	if ( 0 == strcmp ( cap.name, "3Com HomeConnect Camera" ) ) {
		/*
		 * this one actually captures at an odd size, and we have
		 * to do some special magic to get the kernel driver to
		 * scale it properly to something "normal". 
		 */
		int procfd;
		char * devicebase = g_basename ( device );
		char * procfile = 
			g_strdup_printf ( "/proc/3comhc/%s", devicebase );
		procfd = open ( procfile, O_WRONLY );
		if ( procfd < 0 ) {
			perror ( procfile );
			/* eh...  guess we'll just have to live with it. */
		} else {
			/* 
			 * ask the driver to scale the full-size image
			 * for us.
			 */
			write ( procfd, "width=512\n", 10 );
			write ( procfd, "height=242\n", 11 );
			write ( procfd, "rescale=1\n", 10 );
			/* automagickally swap the byte order. */
			write ( procfd, "bgr=1\n", 6 );
//			/* and brighten things up a bit. */
//			write ( procfd, "gamma=2.0\n", 10 );
			close ( procfd );
//			win.width = 512;
//			win.height = 242;
//			if ( 0 > ioctl ( fd, VIDIOCSWIN, & win ) ) {
//				perror("VIDIOCSWIN");
//				return NULL;
//			}
			/* update the capture window structure, since we
			 * changed the dims.  should be 320x240 now. */
			if ( 0 > ioctl ( fd, VIDIOCGWIN, & win ) ) {
				perror("VIDIOCGWIN");
				return NULL;
			}
		}
	} else {
		/* any other... */
		win.width = 320;
		win.height = 240;
		if ( 0 > ioctl ( fd, VIDIOCSWIN, & win ) ) {
			perror("VIDIOCSWIN");
			return NULL;
		}
		/* update the capture window structure, since we
		 * changed the dims.  should be 320x240 now. */
		if ( 0 > ioctl ( fd, VIDIOCGWIN, & win ) ) {
			perror("VIDIOCGWIN");
			return NULL;
		}
	}
	
	v4lcam = (V4LCam*) calloc ( 1, sizeof(V4LCam) );

	camera_construct ( CAMERA(v4lcam), fd, cap.name, 
			win.width, win.height, 3 );
	
	CAMERA(v4lcam)->snap = (CameraFunc) v4lcam_snap;
	CAMERA(v4lcam)->destroy = (CameraFunc) v4lcam_destroy;

	/* create controls with initial values from the device.
	 * caller can override to taste. */
	
	CAMERA(v4lcam)->ncontrols = 5;
	CAMERA(v4lcam)->control = 
		(CamControl**) malloc ( 5 * sizeof(CamControl*) );
	
	CAMERA(v4lcam)->control[ V4LCAM_BRIGHTNESS ] = 
		cam_control_new ( CAMCTRL_SLIDER, " Brightness: ",
			(GtkAdjustment*) gtk_adjustment_new ( 
					(gfloat) pic.brightness, 
					0.0, 65535.0, 1.0, 10.0, 100.0 ) );
	CAMERA(v4lcam)->control[ V4LCAM_CONTRAST ] = 
		cam_control_new ( CAMCTRL_SLIDER, " Contrast: ",
			(GtkAdjustment*) gtk_adjustment_new ( 
					(gfloat) pic.contrast, 
					0.0, 65535.0, 1.0, 10.0, 100.0 ) );
	CAMERA(v4lcam)->control[ V4LCAM_HUE ] = 
		cam_control_new ( CAMCTRL_SLIDER, " Hue: ",
			(GtkAdjustment*) gtk_adjustment_new ( 
					(gfloat) pic.hue, 
					0.0, 65535.0, 1.0, 10.0, 100.0 ) );
	CAMERA(v4lcam)->control[ V4LCAM_COLOR ] = 
		cam_control_new ( CAMCTRL_SLIDER, " Color: ",
			(GtkAdjustment*) gtk_adjustment_new ( 
					(gfloat) pic.colour, 
					0.0, 65535.0, 1.0, 10.0, 100.0 ) );
	CAMERA(v4lcam)->control[ V4LCAM_WHITENESS ] = 
		cam_control_new ( CAMCTRL_SLIDER, " Whiteness: ",
			(GtkAdjustment*) gtk_adjustment_new ( 
					(gfloat) pic.whiteness, 
					0.0, 65535.0, 1.0, 10.0, 100.0 ) );
	
	return CAMERA(v4lcam);
}

int v4lcam_snap ( Camera * cam )
{
	int i;
	byte tmp;
	int buflen = cam->width * cam->height * cam->depth;
	
	/* update the device's settings */
/* FIXME -- it's rather wasteful to set the params *every* time, no?
 *          or is it no big deal?  wouldn't be that hard to cache them...
 */
	struct video_picture pic;
	if ( 0 > ioctl ( cam->handle, VIDIOCGPICT, & pic ) ) {
		perror("VIDIOCGPIC");
		return 0;
	}

	pic.palette    = VIDEO_PALETTE_RGB24;
	pic.depth      = 24;
	pic.brightness = (int) camera_control_value (cam, V4LCAM_BRIGHTNESS);
	pic.contrast   = (int) camera_control_value (cam, V4LCAM_CONTRAST);
	pic.hue        = (int) camera_control_value (cam, V4LCAM_HUE);
	pic.colour     = (int) camera_control_value (cam, V4LCAM_COLOR);
	pic.whiteness  = (int) camera_control_value (cam, V4LCAM_WHITENESS);
	if ( 0 > ioctl ( cam->handle, VIDIOCSPICT, & pic ) ) {
		perror ( "VIDIOCSPICT" );
		return 0;
	}

	if ( 0 > read ( cam->handle, cam->image, buflen ) ) {
		perror ( "read" );
		return 0;
	}

	/* RGB24 is actually BGR... */
	for ( i = 0 ; i < buflen ; i += cam->depth ) {
		tmp = cam->image[i];
		cam->image[i] = cam->image[i+2];
		cam->image[i+2] = tmp;
	}
	
	return 1;
}

void v4lcam_destroy ( Camera * cam )
{
	close ( cam->handle );
}

/****************************************************************************/
