/*
 * muppetcam - uploading webcam viewer
 * by muppet <scott@asofyet.org>, 20-21 june 2002
 *
 * vicam.c - MuppetCam driver for the 3Com HomeConnect USB WebCam.
 *           This is a subclass of Camera.
 *
 *           This is a user-space driver which uses the viutility.[hc] code
 *           from http://homeconnectusb.sourceforge.net and libusb to talk
 *           directly to the camera.
 *           
 *           The camera actually captures 512x242 images, which have an odd
 *           stretched look to them.  This code tries to account for that
 *           by scaling it back to 320x240.
 */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#include "camera.h"
#include "vicam.h"
#include "viutility.h"


typedef struct {
	Camera   cam;
	byte   * capturebuf;
	byte   * decodebuf;
	int      speed;
	int      gain;
	int      realwidth;
	int      realheight;
} ViCam;

#define VICAM(cam) ((ViCam*)cam)

static int  vicam_snap    ( ViCam * cam );
static void vicam_destroy ( ViCam * cam );




Camera * vicam_new ( int handle )
{
	char * name;
	ViCam * vicam;
	
	if ( vicam_claim ( handle ) < 1 ) {
		fprintf ( stderr, "Unable to claim access to camera %d: %s\n", 
				0, strerror(errno) );
		return NULL;
	}
	
	vicam = (ViCam*) calloc ( 1, sizeof ( ViCam ) );

	name = g_strdup_printf ( "ViCam/3Com HomeConnectUSB (%d)", handle );
	camera_construct ( CAMERA(vicam), handle, name, 320, 240, 3 );
	free ( name );

	CAMERA(vicam)->snap    = (CameraFunc) vicam_snap;
	CAMERA(vicam)->destroy = (CameraFunc) vicam_destroy;

	vicam->speed      = 60;
	vicam->gain       = 1;
	vicam->realwidth  = 512;
	vicam->realheight = 242;
	vicam->capturebuf = (byte*) malloc ( 512 * 242 + 128 );
	vicam->decodebuf  = (byte*) malloc ( 512 * 242 * 3 );

	CAMERA(vicam)->ncontrols = 2;
	CAMERA(vicam)->control = 
		(CamControl**) malloc ( 2 * sizeof(CamControl*) );

	CAMERA(vicam)->control[ VICAM_SPEED ] = 
		cam_control_new ( CAMCTRL_SLIDER, "Shutter Speed (1/n)",
			(GtkAdjustment*) gtk_adjustment_new ( 
					(gfloat) vicam->speed, 
					1.0, 3500.0, 1.0, 10.0, 100.0 ) );
	CAMERA(vicam)->control[ VICAM_GAIN ] = 
		cam_control_new ( CAMCTRL_SLIDER, "Gain (Brightness)",
			(GtkAdjustment*) gtk_adjustment_new ( 
					(gfloat) vicam->gain, 
					1.0, 255.0, 1.0, 10.0, 100.0 ) );
	
	return CAMERA(vicam);
}

int vicam_snap ( ViCam * cam )
{
	int srcrowstride, dstrowstride;
	byte * src, * srcrow, * dst, * dstrow;
	int i, j;

	cam->speed = (int) camera_control_value(CAMERA(cam), VICAM_SPEED);
	cam->gain  = (int) camera_control_value(CAMERA(cam), VICAM_GAIN );
	
	vicam_acquire ( CAMERA(cam)->handle, 
			cam->capturebuf, 
			cam->gain,
			cam->speed,
			cam->realwidth,
			cam->realheight,
			camera_is_color ( CAMERA(cam) ) );

	if ( ! camera_is_color ( CAMERA(cam) ) )
		return 1;

	/* decode the color values and scale from the device's stretched wide 
	 * and short weirdness to something normal. */
	
	srcrowstride = cam->realwidth * 3;
	dstrowstride = CAMERA(cam)->width * 3;
	
	vicam_decode_color ( cam->capturebuf,
			     cam->decodebuf,
			     cam->realwidth,
			     cam->realheight );

	/* now scale from wide to normal. */
	for ( i = 0 ; i < CAMERA(cam)->height ; i++ ) {
		srcrow = cam->decodebuf + i * srcrowstride;
		dstrow = CAMERA(cam)->image + i * dstrowstride;
#ifdef ROUGH
		/* select a single pixel to sample by floating point index
		 * calculations. */
		for ( j = 0 ; j < CAMERA(cam)->width ; j++ ) {
			int q = 3 * (int) (j * xscale);
			dstrow[j*3+0] = srcrow[q+0];
			dstrow[j*3+1] = srcrow[q+1];
			dstrow[j*3+2] = srcrow[q+2];
		}
#else
		/* blend fractional hits for a smoother image.
		 * 512 / 320 = 1.6.... that's not going to work very well.
		 * instead, lets just assume 1.5, which means we'll get
		 * some 480 of the 512 pixels of the original image.
		 * then each pair of output corresponds to a triplet on
		 * input, and we'll blend them 2 to 1. */
		src = srcrow;
		dst = dstrow;
		for ( j = 0 ; j < CAMERA(cam)->width ; j+=2 ) {
			dst[0*3+0] = (byte)((2.0*src[0*3+0] + src[1*3+0])/3.0);
			dst[0*3+1] = (byte)((2.0*src[0*3+1] + src[1*3+1])/3.0);
			dst[0*3+2] = (byte)((2.0*src[0*3+2] + src[1*3+2])/3.0);
			
			dst[1*3+0] = (byte)((src[1*3+0] + 2.0*src[2*3+0])/3.0);
			dst[1*3+1] = (byte)((src[1*3+1] + 2.0*src[2*3+1])/3.0);
			dst[1*3+2] = (byte)((src[1*3+2] + 2.0*src[2*3+2])/3.0);
	
			dst += 6; // 3 * 2;
			src += 9; // 3 * 3;
		}
#endif
	}
	
	return 1;
}

void vicam_destroy ( ViCam * cam )
{
	free ( cam->capturebuf );
	free ( cam->decodebuf );
	vicam_release ( 0 );
}


