/* GPL2 boilerplate goes here
 *
 * Bunches of useful functions for writing software to use the vicam
 * with libusb.  Requires libusb 0.1.3 or later.
 * 
 * Copyright 2001 Joseph Burks
 *
 * USB interface code based on vicam.c Copyright 2001 Christopher Cheney
 * Color decode based on perl source Copyright 2001 Andy Armstrong
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <usb.h>
#include <string.h>

#include "vicamurbs.h"
#include "viutility.h"

#define CTRL_TIMEOUT 				500
#define CHROMA_TWEAK					0

static void vicam_add_cam(struct usb_device *newcam);
static void vicam_send_firmware(int camId);
static void vicam_set_power( int camId, int state );


struct vicam_device device[MAX_CAMS];
int vicam_count = 0;
int homeconnect_lite_count = 0;
int dbg_level = 0;

void vicam_set_debug( int level )
{
	dbg_level = level;
}

int vicam_usb_init()
{
	struct usb_bus * bus;
	struct usb_device * dev;
	
	usb_init();
	usb_find_busses();
	usb_find_devices();

	for (bus=usb_busses;bus;bus=bus->next)
	{
		for (dev=bus->devices;dev;dev=dev->next)
		{
			if (dev->descriptor.idVendor == USB_VICAM_VENDOR_ID
				&& dev->descriptor.idProduct == USB_VICAM_PRODUCT_ID)
			{
				if ( vicam_count < MAX_CAMS )
				{
					vicam_add_cam(dev);
				}
			}
			// Check for homeconnect lite.  This software doesn't work
			// with that hardware so finding one of these is probably bad.
			else if ( dev->descriptor.idVendor == USB_3COM_LITE_VENDOR_ID
					&& dev->descriptor.idProduct == USB_3COM_LITE_PRODUCT_ID)
			{
				homeconnect_lite_count++;
			}
		}
	}

	// TODO: Remove nested ternary ops with a nice if/else.
	return vicam_count > 0 ? vicam_count : homeconnect_lite_count > 0 ? -ERR_LITE_NOT_SUPPORTED : -ERR_NO_VICAMS;
}

int vicam_get_cam_count()
{
	return vicam_count;
}

static void vicam_add_cam(struct usb_device *newcam)
{
	device[vicam_count].usbdev = newcam;
	device[vicam_count].cam_id = vicam_count;
	device[vicam_count].claimed = 0;
	vicam_count++;
}

int vicam_claim( int camId )
{
	usb_dev_handle *st_handle = usb_open(device[camId].usbdev);

	if (st_handle == NULL) {
		fprintf(stderr,"usb_open: %s\n",usb_strerror());
		return -1;
	}

	if (usb_set_configuration(st_handle,1) < 0) {
		fprintf(stderr,"usb_set_configuration 1: %s\n",usb_strerror());
		return -1;
	}

	if (usb_claim_interface(st_handle,0) <0) {
		fprintf(stderr,"usb_claim_interface: %s\n",usb_strerror());
		return -1;
	}

	device[camId].handle = st_handle;
	device[camId].claimed = 1;

	vicam_send_firmware(camId);
	vicam_set_power(camId,1);
	vicam_set_led(camId,3);
	
	return 1;
}

int vicam_release( int camId )
{
	if (usb_release_interface(device[camId].handle,0) < 0)
		return -1;

	vicam_set_power(camId,0);
	
	usb_close(device[camId].handle);
	device[camId].claimed = 0;

	return 1;
}	

static void vicam_send_firmware(int camId)
{
	usb_dev_handle *st_handle = device[camId].handle;
	long n;

	n=usb_control_msg(st_handle,0x41,0xff,0,0,
		firmware1,sizeof(firmware1),CTRL_TIMEOUT);
	n=usb_control_msg(st_handle,0x41,0xff,0,0,
		findex1,sizeof(findex1),CTRL_TIMEOUT);

	n=usb_control_msg(st_handle,0x41,0xff,0,0,
		fsetup,sizeof(fsetup),CTRL_TIMEOUT);
	n=usb_control_msg(st_handle,0x41,0xff,0,0,
		firmware2,sizeof(firmware2),CTRL_TIMEOUT);
	n=usb_control_msg(st_handle,0x41,0xff,0,0,
		findex2,sizeof(findex2),CTRL_TIMEOUT);
	n=usb_control_msg(st_handle,0x41,0xff,0,0,
		fsetup,sizeof(fsetup),CTRL_TIMEOUT);
}

static void vicam_set_power( int camId, int state )
{
	usb_control_msg(device[camId].handle,0x41,0x50,state,0,
		NULL,0x00,CTRL_TIMEOUT);
}

long vicam_set_led( int camId, int state )
{
	return usb_control_msg(device[camId].handle,0x41,0x55,state,0,
		NULL,0x00,CTRL_TIMEOUT);
}

void vicam_acquire( int camId, char *buf, char gain, short shutterFraction,
		short width, short height, int color)
{
	unsigned long realShutter;
	unsigned char request[16];
	long n;
	usb_dev_handle *st_handle = device[camId].handle;

	memset(request,0,16);
	request[0] = gain; // 0 = 0% gain, FF = 100% gain

	if ( color )
	{
		request[1] = 0x00;
	}
	else
	{
		request[1] = 0x05; // 512x242 black & white image
	}

	request[2] = 0x90; // the function of these two bytes
	request[3] = 0x07; // is not yet understood
	
	if ( shutterFraction > 60 )
	{
		// Short exposure
		realShutter = ((-15631900 / shutterFraction)+260533)/1000;
		request[4] = realShutter & 0xFF; 
		request[5] = (realShutter >> 8) & 0xFF;
		request[6] = 0x03;
		request[7] = 0x01;
	}
	else
	{
		// Long exposure
		realShutter = 15600/shutterFraction - 1;
		request[4] = 0;
		request[5] = 0;
		request[6] = realShutter & 0xFF;
		request[7] = realShutter >> 8;
	}

	// Per John Markus Bjrndalen, byte at index 8 causes problems if it isn't 0
	request[8] = 0;
	
	// bytes 9-15 do not seem to affect exposure or image quality
	if ( dbg_level > 0 )
	{
		fprintf(stderr,"debug: Image acquire packet: ");
		for ( n = 0; n < 16; n++ )
		{
			fprintf(stderr," %02x",request[n]);
		}
		fprintf(stderr,"\n");
	}

	n = usb_control_msg(st_handle,0x41,0x51,0x80,0,
		request,0x10,CTRL_TIMEOUT);

	if ( n == -1 )
	{
		fprintf(stderr,"usb_control_msg: %s\n",usb_strerror());
    return;
	}

	n = usb_bulk_read(st_handle,0x81,buf,
		0x1e480,CTRL_TIMEOUT);

	if ( n == -1 )
		fprintf(stderr,"usb_bulk_read: %s\n",usb_strerror());

	return;
}

void vicam_decode_color( char *data, char *rgb, int width, int height )
{
  /**
  * Most of this code was copied directly from the perl source.
  */
    int x,y;
    int addr,dx,dy,orig;
    int sp[4] = {0,0,0,0};
    double luma,cx,cy,rc,gc,bc;
#if CHROMA_TWEAK
    double chroma = 1.0;
#endif
    int red,green,blue;
	 int offset;

    //# Now scan the array to build the bitmap
    for(y = 0; y < height; y++)
    {
      for(x = 0; x < width; x++)
      {
        addr = y * width + x + 64;
        dx = (x < width-1) ? 1 : -1;
        dy = (y < height-1) ? width : -width;
        orig = ((x & 1) << 1) | ((y & 1) << 0);
        sp[orig ^ 0] = data[addr] & 0xFF;       // 00
        sp[orig ^ 1] = data[addr+dy] & 0xFF;      // 01
        sp[orig ^ 2] = data[addr+dx] & 0xFF;      // 10
        sp[orig ^ 3] = data[addr+dx+dy] & 0xFF; // 11

        //# Chroma
        cx = sp[3] - sp[1];
        cy = sp[2] - sp[0];
        rc = cy + cx * 0.500;
        gc = cy * -0.500 + cx * 0.866;
        bc = cy * -0.500 - cx * 0.866;

        //# Calc luma (0-255)
		  // I'm not sure what was wrong with the old luma calculation, however
		  // this change works better (doubling horizontal resolution)
		  sp[0] = data[addr] & 0xFF;
		  sp[1] = data[addr+1] & 0xFF;
		  luma = (int)((sp[0] + sp[1])/2);

#if CHROMA_TWEAK
		  red = (int)(luma + rc * chroma);
        green = (int)(luma + gc * chroma);
        blue  = (int)(luma + bc * chroma);
#else
		  red = (int)(luma + rc);
        green = (int)(luma + gc);
        blue  = (int)(luma + bc);
#endif

        red   = (red   > 255) ? 255 : ((red   < 0) ? 0 : red  );
        green = (green > 255) ? 255 : ((green < 0) ? 0 : green);
        blue  = (blue  > 255) ? 255 : ((blue  < 0) ? 0 : blue );

		  offset = y*width*3+x*3;
		  rgb[offset+0] = red;
		  rgb[offset+1] = green;
		  rgb[offset+2] = blue;
      }
    }
	return;
}

